"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.runTasks = void 0;
const tslib_1 = require("tslib");
const chalk_1 = tslib_1.__importDefault(require("chalk"));
const logger = tslib_1.__importStar(require("../../logger"));
const runTasks = (options) => {
    const tasks = [
        {
            // requires are put here for a reason, it is to delay code execution until needed. That's because one task
            // might change files that another task relies on. If all files are required beforehand, consecutive tasks will
            // not see file changes and thus, component generation becomes falsy
            task: () => require('./tasks/verify-working-directory').verifyWorkingDirectory,
            skipped: options.force,
            message: 'Verify clean working directory',
        },
        {
            task: () => require('./tasks/copy-templates').copyTemplates,
            message: 'Copy templates',
        },
        {
            task: () => require('./tasks/run-codemods').runCodemods,
            message: 'Run codemods',
            skipped: !options.codemods,
        },
    ];
    logger.info(`Generating ${chalk_1.default.cyan(`<${options.ComponentName}/>`)} component...`);
    return tasks
        .filter(({ skipped }) => !skipped)
        .reduce((promise, { task, message = '' }) => promise.then(() => {
        const spinner = logger.spinner(message);
        return task()(options)
            .then(() => {
            spinner.stop();
            logger.success(`Done: ${message}`);
        })
            .catch((e) => {
            spinner.stop();
            logger.error(`Failed: ${message} ${e}`);
            return Promise.resolve();
        });
    }), Promise.resolve())
        .then(() => logger.success(`${chalk_1.default.cyan(`<${options.ComponentName}/>`)} generated successfully!`))
        .catch((e) => {
        logger.error(e);
        process.exit(1);
    });
};
exports.runTasks = runTasks;
//# sourceMappingURL=run-tasks.js.map