"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.resolvePlugin = exports.executePlugin = void 0;
const tslib_1 = require("tslib");
const path_1 = tslib_1.__importDefault(require("path"));
const promises_1 = tslib_1.__importDefault(require("fs/promises"));
const errors_1 = require("../../errors");
const template_1 = require("./template");
const path_resolver_1 = require("./path-resolver");
const resolve_export_1 = require("./resolve-export");
const string_utils_1 = require("./string-utils");
const cleanFolder = (cwd) => (folderPath) => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
    const absoluteFolderPath = path_1.default.resolve(folderPath);
    if (absoluteFolderPath.length < cwd.length) {
        throw new errors_1.WufError({
            kind: "UserError" /* UserError */,
            name: 'CleanFolderError',
            message: `Declining to remove file outside of current working directory. This folder will not be cleaned: "${absoluteFolderPath}"`,
        });
    }
    try {
        yield promises_1.default.stat(folderPath);
    }
    catch (e) {
        const error = e;
        if (error.message && error.message.includes('ENOENT')) {
            // returning void when folder does not exist is expected behaviour
            return;
        }
        else {
            throw new errors_1.WufError({
                kind: "SystemError" /* SystemError */,
                name: 'CleanFolderError',
                message: `Unexpected error occured while trying to clean folder at ${absoluteFolderPath}`,
                error,
            });
        }
    }
    const files = yield promises_1.default.readdir(folderPath, { encoding: 'utf8' });
    for (const file of files) {
        const removablePath = path_1.default.join(folderPath, file);
        try {
            yield promises_1.default.rm(removablePath, { recursive: true, force: true });
        }
        catch (error) {
            throw new errors_1.WufError({
                kind: "SystemError" /* SystemError */,
                name: 'CleanFolderError',
                message: `Error occurred while trying to remove file ${removablePath}`,
                error,
            });
        }
    }
});
const chain = ({ plugins, data, api, }) => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
    return plugins.reduce((data, fn) => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
        try {
            const previousData = yield data;
            const newData = yield fn(previousData, api);
            return Object.assign(Object.assign({}, previousData), newData);
        }
        catch (error) {
            throw new errors_1.WufError({
                kind: "UserError" /* UserError */,
                name: 'PluginError',
                message: 'Plugin is failing at runtime',
                error,
            });
        }
    }), data);
});
const executePlugin = ({ requirePath, data, api }) => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
    const pluginRaw = (0, errors_1.resolveOrThrow)(() => require(requirePath), (error) => new errors_1.WufError({
        name: 'PluginError',
        kind: "SystemError" /* SystemError */,
        message: `Unable to require plugin at "${requirePath}"`,
        error,
    }));
    const plugin = (0, resolve_export_1.resolveExport)(pluginRaw);
    if (typeof plugin === 'function') {
        const pluginOutput = yield plugin(data, api);
        return Object.assign(Object.assign({}, data), pluginOutput);
    }
    else if (Array.isArray(plugin)) {
        return Object.assign(Object.assign({}, data), (yield chain({
            plugins: plugin,
            data,
            api,
        })));
    }
    return plugin;
});
exports.executePlugin = executePlugin;
const resolvePlugin = ({ options, components }) => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
    const pathResolve = (0, path_resolver_1.pathResolver)(options._process.cwd);
    const pluginPaths = options.plugin.map((rawPluginPath) => {
        const pluginPath = pathResolve(rawPluginPath);
        const relativePath = path_1.default.relative(options._process.cwd, pluginPath);
        const errors = [];
        try {
            return require.resolve(pluginPath);
        }
        catch (e) {
            const error = new errors_1.WufError({
                name: 'PluginError',
                kind: "UserError" /* UserError */,
                message: `Plugin not found at "${relativePath}"`,
                error: e,
            });
            errors.push(error);
        }
        try {
            const internalPluginPath = path_1.default.join('wix-ui-framework', 'plugins', rawPluginPath);
            return require.resolve(internalPluginPath, {
                paths: [options._process.cwd],
            });
        }
        catch (e) {
            const error = new errors_1.WufError({
                name: 'PluginError',
                kind: "UserError" /* UserError */,
                message: `Plugin not found at "${relativePath}"`,
                error: e,
            });
            errors.push(error);
        }
        if (errors.length) {
            throw errors[0];
        }
    });
    const pluginApi = Object.assign({ cwd: options._process.cwd, output: options.output, renderTemplate: template_1.renderTemplate, cleanFolder: cleanFolder(options._process.cwd) }, string_utils_1.stringUtils);
    return yield pluginPaths.reduce((promise, requirePath) => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
        return (0, exports.executePlugin)({
            requirePath,
            data: yield promise,
            api: pluginApi,
        });
    }), Promise.resolve({ components }));
});
exports.resolvePlugin = resolvePlugin;
//# sourceMappingURL=plugin.js.map