"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.update = void 0;
const tslib_1 = require("tslib");
const path_1 = tslib_1.__importDefault(require("path"));
const promises_1 = tslib_1.__importDefault(require("fs/promises"));
const minimatch_1 = tslib_1.__importDefault(require("minimatch"));
const file_exists_1 = require("../../file-exists");
const object_entries_1 = require("../../object-entries");
const fs_to_json_1 = require("../../fs-to-json");
const map_tree_1 = require("../../map-tree");
const get_dirty_components_1 = require("../../get-dirty-components");
const get_changed_files_1 = require("../../get-changed-files");
const pathResolver = (cwd) => (...a) => path_1.default.resolve(cwd, ...a);
const readOutputFile = (path = '') => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
    try {
        const outputRaw = yield promises_1.default.readFile(path, 'utf8');
        return JSON.parse(outputRaw);
    }
    catch (e) {
        return {};
    }
});
const augmentWithDirty = ({ cwd, components, changedFiles }) => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
    const dirtyComponents = yield (0, get_dirty_components_1.getDirtyComponents)({
        rootPath: cwd,
        components,
        changedFiles,
    });
    const output = Object.entries(components).reduce((acc, [componentName, component]) => {
        acc[componentName] = dirtyComponents.includes(componentName)
            ? Object.assign(Object.assign({}, component), { dirty: true }) : component;
        return acc;
    }, {});
    return output;
});
const guards = (unsafeOptions) => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
    const pathResolve = pathResolver(unsafeOptions._process.cwd);
    const options = Object.assign(Object.assign({}, unsafeOptions), { shape: pathResolve(unsafeOptions.shape || '.wuf/required-component-files.json'), components: pathResolve(unsafeOptions.components || 'src/components'), output: pathResolve(unsafeOptions.output || '.wuf/components.json'), maxMismatch: unsafeOptions.maxMismatch || 0, exclude: unsafeOptions.exclude });
    if (unsafeOptions.verboseOutput) {
        options.verboseOutput = pathResolve(options.verboseOutput);
    }
    if (!(yield (0, file_exists_1.fileExists)(options.shape))) {
        throw new Error(`Component structure file does not exist at "${unsafeOptions.shape || options.shape}"`);
    }
    if (!(yield (0, file_exists_1.fileExists)(options.components))) {
        throw new Error(`Cannot read components folder at "${unsafeOptions.components}"`);
    }
    return makeOutput(options);
});
const makeOutput = (options) => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
    const shapeRaw = yield promises_1.default.readFile(options.shape, 'utf8');
    const shape = JSON.parse(shapeRaw);
    const output = yield readOutputFile(options.output);
    const componentsFs = yield (0, fs_to_json_1.fsToJson)({
        cwd: options.components,
        path: '.',
    });
    const analyzedComponents = (0, object_entries_1.objectEntries)(componentsFs)
        .filter(([name, structure]) => [
        // skip non folders at root level
        structure !== '',
        // skip excluded
        options.exclude ? !new RegExp(options.exclude).test(name) : true,
    ].every(Boolean))
        .map(([name, structure]) => {
        // TODO: this should be a placeholder coming from config. There should be support for multiple
        const replaceableName = 'Component';
        const namedShape = (0, map_tree_1.mapTree)(shape, ({ key, value }) => ({
            [key.replace(replaceableName, name)]: value,
        }));
        const missingFiles = [];
        const traverseGlobs = ({ globTree, fsTree, fsPath = '' }) => {
            const globEntries = (0, object_entries_1.objectEntries)(globTree);
            const fsEntries = (0, object_entries_1.objectEntries)(fsTree);
            return globEntries.reduce((row, [globPath, globValue]) => {
                const matchingFsEntries = fsEntries.filter(([entry]) => (0, minimatch_1.default)(entry, globPath));
                if (!matchingFsEntries.length) {
                    missingFiles.push(path_1.default.join(fsPath.replace(/^\//, ''), globPath));
                }
                matchingFsEntries.map(([matchingFsPath, matchingFsValue]) => {
                    if (typeof matchingFsValue !== 'string') {
                        row[matchingFsPath] = traverseGlobs({
                            fsTree: matchingFsValue,
                            globTree: globValue,
                            fsPath: `${fsPath}/${matchingFsPath}`,
                        });
                    }
                    else {
                        row[matchingFsPath] = matchingFsValue;
                    }
                });
                return row;
            }, {});
        };
        traverseGlobs({
            globTree: namedShape,
            fsTree: structure,
        });
        return {
            name,
            structure,
            missingFiles,
            path: path_1.default.relative(options._process.cwd, path_1.default.resolve(options.components, name)),
        };
    });
    const goodComponents = analyzedComponents.filter(({ missingFiles }) => missingFiles.length <= options.maxMismatch);
    const jsonOutput = goodComponents.reduce((components, component) => {
        components[component.name] = Object.assign({ path: component.path }, (component.missingFiles.length
            ? { missingFiles: component.missingFiles }
            : {}));
        return components;
    }, output);
    if (options.verboseOutput) {
        let changedFiles;
        try {
            changedFiles = yield (0, get_changed_files_1.getChangedFiles)({ cwd: options._process.cwd });
        }
        catch (e) {
            changedFiles = [];
        }
        const verboseOutput = yield augmentWithDirty({
            cwd: options._process.cwd,
            components: jsonOutput,
            changedFiles,
        });
        yield promises_1.default.writeFile(options.verboseOutput, JSON.stringify(verboseOutput, null, 2));
    }
    yield promises_1.default.writeFile(options.output, JSON.stringify(jsonOutput, null, 2));
});
exports.update = guards;
//# sourceMappingURL=index.js.map